
from pathlib import Path
from datetime import datetime

from plover_ninja.lessons.make_a_new_friend import MakeANewFriend
from plover_ninja.lessons.need_for_speed import NeedForSpeed
from plover_ninja.lessons.remember_me import RememberMe
from plover_ninja.plugins.ninja_plugin import NinjaPlugin
from plover_ninja import storage


CHORDING_PAUSED_IN_SECONDS = 5


class StrokeEfficiencyLog(NinjaPlugin):

    def __init__(self, engine):
        """
        resumed_chording - beginning of active typing, resets after a noticeable pause
        previous_stroke - literally, when the last stroke was made
        latest_stroke - _this_ stroke that just happened
        """
        super().__init__(engine)
        self.resumed_chording = None
        self.previous_stroke = None
        self.latest_stroke = None
        self.stroke_efficiency_log = storage.StrokeEfficiencyLog()
        self.words_written_this_session = 0

        with open(Path(storage.TEMP_DIR) / 'ninja-stroke-efficiency.txt', 'a', encoding='utf-8') as f:
            f.write(f'Stroke Efficiency Log is up and running! 🎉\n')

        # TODO: Move lesson creation somewhere else?
        # Create a lesson!
        MakeANewFriend().make_lesson()
        RememberMe().make_lesson()
        NeedForSpeed().make_lesson()


    def _log_efficiency_report(self):
        efficiency_map = self.stroke_efficiency_log.get_average_speed_and_frequency_for_stroked_words()

        with open(Path(storage.TEMP_DIR) / 'ninja-stroke-efficiency.txt', 'a', encoding='utf-8') as f:
            f.write('\n----------------------------------\n')
            f.write('New stroke efficiency results! 🎉\n\n')
            for word, _, avg_stroke_duration in efficiency_map:
                f.write(f'{word} => {avg_stroke_duration}\n')
            f.write('\n')


    def on_translated(self, old, new):
        if len(new) == 0:
            return

        word = new[0].word

        if self.resumed_chording is None:
            self.resumed_chording = self.previous_stroke = self.latest_stroke = datetime.now()
            return
        else:
            self.previous_stroke = self.latest_stroke
            self.latest_stroke = datetime.now()

        self.words_written_this_session += 1

        # time in seconds (expressed as a float; includes fractions of a second)
        gap_between_strokes = (self.latest_stroke - self.previous_stroke).total_seconds()

        if gap_between_strokes > CHORDING_PAUSED_IN_SECONDS:
            # can't accurately determine how long this stroke took
            self.resumed_chording = self.latest_stroke
            return

        # record time it took to make last stroke
        self.stroke_efficiency_log.add_stroke(word, gap_between_strokes)
        with open(Path(storage.TEMP_DIR) / 'ninja-stroke-efficiency.txt', 'a', encoding='utf-8') as f:
            f.write(f'Stroked {word} in {gap_between_strokes:.2f}\n')

    def on_stroked(self, stroke):
        pass



#!/usr/bin/env python3

import pathlib

CURR_DIR = pathlib.Path(__file__).parent.absolute()
WORD_FREQUENCY_LIST = pathlib.PurePath(CURR_DIR, 'enwiki-20210820-words-frequency.txt')

def get_word_frequency_list_as_map():
    word_frequency_map = {}
    with open(WORD_FREQUENCY_LIST, 'r') as f:
        lines = f.readlines()
        for line in lines:
            tokens = line.split()
            word_frequency_map[tokens[0]] = tokens[1]
    return word_frequency_map




import logging
import tempfile
from pathlib import Path

from plover_ninja.plugins.ninja_plugin import NinjaPlugin

logger = logging.getLogger(__name__)

TEMP_DIR = tempfile.gettempdir()

class NinjaTest(NinjaPlugin):

    def on_translated(self, old, new):
        logger.info(f'Something was translated! {old} {new}\n')

        with open(Path(TEMP_DIR) / 'ninja.txt', 'a', encoding='utf-8') as ninja_log:
            ninja_log.write(f'Something was translated! {old} {new}\n')

    def on_stroked(self, stroke):
        logger.info(f'Something was stroked! {stroke}\n')

        with open(Path(TEMP_DIR) / 'ninja.txt', 'a', encoding='utf-8') as ninja_log:
            ninja_log.write(f'Something was stroked! {stroke}\n')



#!/usr/bin/env python3

from pathlib import Path
from datetime import date

from plover_ninja.storage import get_rarely_used_words, TEMP_DIR


class RememberMe:
    def __init__(self):
        pass

    # TODO: Make this a class method?
    def make_lesson(self):
        # returns a tuple containing:
        # - count of strokes
        # - word frequency
        # - word
        least_used_words_tuples = get_rarely_used_words()

        longest_word_len = 0
        for _, _, word in least_used_words_tuples:
            word_len = len(word)
            if word_len > longest_word_len:
                longest_word_len = word_len
        word_padding = longest_word_len + 5

        longest_word_id_len = 0
        for _, word_id, _ in least_used_words_tuples:
            word_id_len = len(str(word_id))
            if word_id_len > longest_word_id_len:
                longest_word_id_len = word_id_len
        word_id_padding = longest_word_id_len + 5

        word_list = []
        for stroke_count, word_id, word in least_used_words_tuples:
            text = f'{word:{word_padding}} word #{word_id:<{word_id_padding}} stroke count: {stroke_count}'
            word_list.append(text)
        word_list = '\n'.join(word_list)

        today = date.today()
        file_name_today = today.strftime('%Y%m%d')
        report_text_today = today.strftime('%m/%d/%Y')

        with open(f'{Path(TEMP_DIR) / file_name_today}_remember_me.txt', 'w', encoding='utf-8') as f:
            text = f"""Ninja Lesson
{report_text_today}

You've seen these words, but they're still pretty fresh.
Take some time getting to know them a bit better!

{word_list}

🐦🥋
"""
            f.write(text)



#!/usr/bin/env python3

from pathlib import Path
from datetime import date

from plover_ninja.storage import get_most_common_words_that_have_not_been_used_yet, TEMP_DIR


class MakeANewFriend:
    def __init__(self):
        pass

    # TODO: Make this a class method?
    def make_lesson(self, mini_lesson = False):
        frequency_word_tuples = get_most_common_words_that_have_not_been_used_yet()
        longest_word_len = 0
        for _, word in frequency_word_tuples:
            word_len = len(word)
            if word_len > longest_word_len:
                longest_word_len = word_len
        padding = longest_word_len + 5

        word_list = []
        detailed_word_list = []
        for frequency, word in frequency_word_tuples:
            text = f'{word:{padding}} word #{frequency}'
            detailed_word_list.append(text)
            word_list.append(word)
        word_list = '\n'.join(word_list)
        detailed_word_list = '\n'.join(detailed_word_list)

        today = date.today()
        file_name_today = today.strftime('%Y%m%d')
        report_text_today = today.strftime('%m/%d/%Y')

        with open(f'{Path(TEMP_DIR) / file_name_today}_make_a_new_friend.txt', 'w', encoding='utf-8') as f:
            text = f"""Ninja Lesson
{report_text_today}

Make some new friends today! Here are the most common words
that haven't appeared in your writing yet. Give them a try!

{detailed_word_list}

🐦🥋
"""
            f.write(text)

        if mini_lesson:
            return f"""New Words!
----------
{word_list}
"""
        return text



#!/usr/bin/env python3

import math
import os
import sqlite3
import tempfile
import threading
from datetime import date
from pathlib import Path
from uuid import uuid4

from plover_ninja.wikipedia_word_frequency.word_frequency_list_manager import get_word_frequency_list_as_map

connection = None

HOME = str(Path.home())
TEMP_DIR = tempfile.gettempdir()
DB_DIR = os.path.join(HOME, ".plover_ninja")
DB_FILE = os.path.join(DB_DIR, "ninja.db")  # os.path.join returns a string


def get_connection():
    os.makedirs(DB_DIR, exist_ok=True)
    db_file = Path(DB_FILE)
    if not db_file.is_file():
        db_file.touch()
        db_file.chmod(0o700)  # Octal: 700
    global connection
    if not connection:
        connection = sqlite3.connect(DB_FILE)
    return connection


def close_connection():
    global connection
    if connection:
        connection.close()
        connection = None


def today():
    d = date.today()
    return d.strftime("%Y-%m-%d")


class SettingsManager:
    def __init__(self):
        self.connection = get_connection()
        self._ensure_tables_exists()

    def _ensure_tables_exists(self):
        cur = self.connection.cursor()
        try:
            cur.execute('SELECT * FROM Settings LIMIT 1')
            cur.fetchone()
        except sqlite3.OperationalError:
            cur.execute("""CREATE TABLE Settings (setting_id INTEGER PRIMARY KEY,
                                                  name TEXT,
                                                  value TEXT)""")
            self.connection.commit()

    def get_setting(self, name):
        """Returns setting if found, otherwise returns `None`"""
        cur = self.connection.cursor()
        t = (name,)
        cur.execute("""SELECT value FROM Settings
                       WHERE name = ?""", t)
        res = cur.fetchone()
        if res is None:
            return None
        return res[0]

    def setting_exists(self, name):
        return self.get_setting(name) is not None

    def set_setting(self, name, value):
        cur = self.connection.cursor()
        if self.setting_exists(name):
            t = (value, name)
            cur.execute("""UPDATE Settings
                           SET value = ?
                           WHERE name = ?""", t)
        else:
            t = (name, value)
            cur.execute("""INSERT INTO Settings(name, value)
                           VALUES (?, ?)""", t)
        self.connection.commit()


class StrokeEfficiencyLog:
    def __init__(self):
        self.connection = get_connection()
        self.strokes_since_average_recomputed = 0
        self.avg_stroke_recalculation_thread = None
        StrokeEfficiencyLogInitializer().initialize()

    def add_stroke(self, word, stroke_duration):
        """Note how long a stroke took.

        Returns True if entry was added successfully, False
        otherwise."""
        # Before processing stroke, see if any existing
        # threads for recalculating stroke duration averages
        # have finished
        t = self.avg_stroke_recalculation_thread
        if t is not None and not t.is_alive():
            self.avg_stroke_recalculation_thread = None

        if word is None:
            return

        # Our corpus of words that we used to form the words
        # table is case-insensitive, so the strokes we log
        # will be, too
        word = word.lower()

        cur = self.connection.cursor()
        # sqlite date functions:
        # https://sqlite.org/lang_datefunc.html
        #
        # hex() is a helpful sqlite function for
        # viewing the uuid4 primary key value
        # used for stroke_uuid
        # https://sqlite.org/lang_corefunc.html#hex
        primary_key = uuid4().bytes
        t = (primary_key, stroke_duration, word)
        cur.execute("""INSERT INTO Strokes(stroke_uuid, word_id, date, stroke_duration)
                        SELECT ?, Words.word_id, DATE('now', 'localtime'), ?
                        FROM Words
                        WHERE Words.word = ? LIMIT 1""", t)
        # indicate that average stroke duration needs to
        # be recalculated for this word
        t = (word,)
        cur.execute("""UPDATE Words
                        SET average_stroke_duration_dirty_flag = 1
                        WHERE word = ?""", t)
        self.connection.commit()

        # periodically recompute average stroke duration
        self.strokes_since_average_recomputed += 1

        if self.strokes_since_average_recomputed > 100:
            self.strokes_since_average_recomputed = 0
            # ensure that any existing threads that were
            # started to recompute average stroke duration
            # have finished executing. If they have not, raise
            # an exception as this may be an indication
            # that the thread's task has run into an issue.
            if self.avg_stroke_recalculation_thread is not None:
                error_msg = 'Previous process that was recomputing ' + \
                    'the average stroke duration never closed.\n' + \
                    f'{self.avg_stroke_recalculation_proc.exitcode}'
                # crash hard
                raise Exception(error_msg)
            self.avg_stroke_recalculation_thread = update_average_stroke_duration(blocking=False)
        return True

    def get_average_speed_and_frequency_for_stroked_words(self):
        cur = self.connection.cursor()
        cur.execute("""SELECT Words.word,
                              Words.frequency,
                              AVG(Strokes.stroke_duration) AS AverageDuration
                         FROM Words
                         JOIN Strokes ON Words.word_id = Strokes.word_id
                         GROUP BY Words.word
                         ORDER BY Words.frequency DESC
                    """)
        rows = cur.fetchall()
        return rows


class StrokeEfficiencyLogInitializer:
    # TODO: Make these class methods
    def __init__(self):
        self.connection = get_connection()

    def initialize(self):
        self._ensure_tables_exist()

    def _ensure_tables_exist(self):
        cur = self.connection.cursor()
        # Words table
        try:
            cur.execute('SELECT * FROM Words LIMIT 1')
            cur.fetchone()
        except sqlite3.OperationalError:
            # sqlite does not have a separate boolean data type;
            # INTEGER is used instead
            # https://www.sqlite.org/datatype3.html#boolean_datatype
            cur.execute("""CREATE TABLE Words (word_id INTEGER PRIMARY KEY,
                                               word TEXT,
                                               frequency INTEGER,
                                               practice_weight REAL,
                                               average_stroke_duration REAL DEFAULT NULL,
                                               average_stroke_duration_dirty_flag INTEGER DEFAULT NULL)""")
            cur.execute('CREATE INDEX WordIdIndex ON Words(word_id)')
            cur.execute('CREATE INDEX WordIndex ON Words(word)')
            self.connection.commit()

            # since table didn't exist, need to
            # populate table
            self._populate_words_table()

        # Strokes table
        try:
            cur.execute('SELECT * FROM Strokes LIMIT 1')
            cur.fetchone()
        except sqlite3.OperationalError:
            # Create the Strokes table using `WITHOUT ROWID`
            # in order to prevent the rows from being
            # listed in the order in which they were added
            # by a SELECT statement.
            #
            # https://www.sqlite.org/withoutrowid.html
            cur.execute("""CREATE TABLE Strokes (stroke_uuid BLOB PRIMARY KEY,
                                                 word_id INT,
                                                 date INTEGER,
                                                 stroke_duration REAL,
                                                 FOREIGN KEY(word_id) REFERENCES Words(word_id))
                                                WITHOUT ROWID""")
            cur.execute('CREATE INDEX StrokedWordIndex ON Strokes(word_id)')
            self.connection.commit()

    def _populate_words_table(self):
        def get_practice_weight(t):
            t = int(t)
            return math.log(t, 1.88265334)

        # note that this returns only lower-case words
        word_frequency_map = get_word_frequency_list_as_map()
        cur = self.connection.cursor()
        for word, frequency in word_frequency_map.items():
            t = (word, frequency, get_practice_weight(frequency))
            cur.execute('INSERT INTO Words (word, frequency, practice_weight) VALUES (?, ?, ?)', t)
        self.connection.commit()


###################################################
# Lessons

def get_most_common_words_that_have_not_been_used_yet(num_words=5):
    connection = get_connection()
    cur = connection.cursor()

    t = (num_words, )
    cur.execute("""SELECT word_id, word FROM Words
                   WHERE word_id NOT IN
                     (SELECT word_id FROM Strokes)
                   ORDER BY word_id ASC
                   LIMIT ?""", t)
    words = cur.fetchall()
    return words


def get_rarely_used_words(num_words=10, threshold=100):
    """Returns list of words that have been used less than
    `threshold` number of times. Scans words according
    to their general frequency of use.

    Returns list of tuples where each tuple contains:
    - Count of strokes
    - Word frequency
    - Word
    """
    connection = get_connection()
    cur = connection.cursor()

    t = (num_words, )
    cur.execute("""SELECT COUNT(Strokes.stroke_uuid) AS StrokeCount,
                        Words.word_id,
                        Words.word
                        FROM Strokes
                        JOIN Words ON Strokes.word_id = Words.word_id
                        GROUP BY Strokes.word_id
                        ORDER BY Words.frequency DESC
                        LIMIT ?
                """, t)
    words = cur.fetchall()
    return words


def get_slowest_stroked_words(num_words=10):
    connection = get_connection()
    cur = connection.cursor()

    # update average stroke duration for
    # any words where the `average_stroke_duration_dirty_flag`
    # is TRUE
    update_average_stroke_duration()

    t = (num_words,)
    cur.execute("""SELECT word_id,
                          word,
                          practice_weight,
                          average_stroke_duration,
                          average_stroke_duration * Words.practice_weight AS WeightedDuration
                          FROM Words
                          WHERE NOT average_stroke_duration IS NULL
                          ORDER BY WeightedDuration DESC
                          LIMIT ?
                """, t)
    words = cur.fetchall()
    return words


def _update_average_stroke_duration(db_file, name='no_name'):
    connection = sqlite3.connect(db_file)
    cur = connection.cursor()

    # get list of words where the average needs
    # to be recomputed
    cur.execute("""SELECT word_id
                   FROM WORDS
                   WHERE average_stroke_duration_dirty_flag = 1""")
    res = cur.fetchall()
    word_ids = tuple(t[0] for t in res)

    # calculate average stroke duration
    # There's got to be a better way to SELECT from
    # a dynamic list
    # https://stackoverflow.com/a/5766293
    cur.execute(f"""SELECT word_id, AVG(stroke_duration)
                    FROM Strokes
                    WHERE word_id IN ({','.join(['?']*len(word_ids))})
                    GROUP BY word_id
                """, word_ids)
    res = cur.fetchall()

    # save new averages
    for word_id, avg in res:
        t = (avg, word_id)
        cur.execute("""UPDATE Words
                        SET average_stroke_duration = ?,
                            average_stroke_duration_dirty_flag = 0
                        WHERE word_id = ?""", t)
    # commit results
    connection.commit()
    connection.close()


def update_average_stroke_duration(blocking=True):
    """If blocking is True, will update averages synchronously.

    If `blocking` is False, will return a Thread object that
    has been started.
    """
    if blocking:
        _update_average_stroke_duration(DB_FILE)
        return None

    t = threading.Thread(target=_update_average_stroke_duration, args=(DB_FILE, uuid4()))
    t.start()
    return t


def get_daily_number_of_strokes():
    connection = sqlite3.connect(DB_FILE)
    cur = connection.cursor()

    # get list of words where the average needs
    # to be recomputed
    cur.execute("""SELECT date, count(stroke_uuid)
                   FROM Strokes
                   GROUP BY date
                   ORDER BY Date Desc""")
    res = cur.fetchall()
    return {k: v for k, v in res}



class NinjaPlugin:
    def __init__(self, engine):
        pass

    def on_translated(self, old, new):
        # Do nothing by default
        pass

    def on_stroked(self, stroke):
        # Do nothing by default
        pass



from datetime import date, timedelta

from plover_ninja.plugins.ninja_plugin import NinjaPlugin
from plover_ninja.lessons.need_for_speed import NeedForSpeed
from plover_ninja.lessons.make_a_new_friend import MakeANewFriend
from plover_ninja.storage import get_daily_number_of_strokes


class NinjaRepl(NinjaPlugin):
    def __init__(self, engine):
        self.engine = engine
        self.listener_manager = ListenerManager()

        # Load initial state
        WelcomeToNinjaTraining(self.engine, self.listener_manager).load()

    def on_translated(self, old, new):
        if len(new) == 0:
            return
        new_word = new[0].word

        if new_word is None:
            return
        new_word = new_word.lower()

        event = {'new_word': new_word}
        self.listener_manager.send_event_to_listeners(event)


class ListenerManager:
    def __init__(self):
        self.listeners = {}
        self.permanent_listeners = {}

    def clear_listeners(self):
        self.listeners = {}

    def add_listener(self, listener_name, listener):
        self.listeners[listener_name] = listener

    def add_permanent_listener(self, listener_name, listener):
        self.permanent_listeners[listener_name] = listener

    def remove_listener(self, listener_name):
        del self.listeners[listener_name]

    def send_event_to_listeners(self, event_dict):
        for listener in self.permanent_listeners.values():
            listener.on_event_occurred(event_dict)

        for listener in self.listeners.values():
            listener.on_event_occurred(event_dict)


##############################
# Listeners

class WaitForPhrase:
    """WaitForPhrase waits for a *case-sensitive* phrase!"""
    def __init__(self, phrase, callback):
        self.phrase = phrase
        self.received_words = []
        self.callback = callback

    def on_event_occurred(self, event):
        if 'new_word' not in event:
            return
        word = event['new_word']

        self.received_words.append(word)

        if len(self.received_words) > len(self.phrase):
            self.received_words = self.received_words[1:]

        if self.received_words == self.phrase:
            self.callback.run()


##############################
# Callbacks

class GreetingCallback:
    def __init__(self, engine, mini_lesson=True):
        self.engine = engine
        self.mini_lesson = mini_lesson

    def run(self):
        #self.engine._send_string(f'Howdy!🐦🥋\n')

        lesson_text = MakeANewFriend().make_lesson(mini_lesson=self.mini_lesson)
        self.engine._send_string(lesson_text)

        self.engine._send_string('\n')

        lesson_text = NeedForSpeed().make_lesson(mini_lesson=self.mini_lesson)
        self.engine._send_string(lesson_text)

class StatsCallback:
    def __init__(self, engine, days=30):
        self.engine = engine
        self.days = days

    def run(self):
        stats_text = self.get_stats_content(self.days)
        for line in stats_text:
            self.engine._send_string(f'{line}\n')

    def get_stats_content(self, days):
        date_to_stroke_count_map = get_daily_number_of_strokes()

        # build custom list that:
        # - has entries for each day (including 0 if needed)
        # - only goes back `days`
        date_to_stroke_stats_list = []

        today = date.today()
        for i in range(0, days):
            curr_date = today - timedelta(days=i)
            curr_date_str = curr_date.strftime('%Y-%m-%d')

            num_strokes = date_to_stroke_count_map.get(curr_date_str, 0)
            date_to_stroke_stats_list.append((curr_date_str, num_strokes))

        max_stroke_count = max(strokes for _, strokes in date_to_stroke_stats_list)
        stroke_count_width = len(str(max_stroke_count))

        lines = []
        lines.append(f'{"Date":<13} {"# Words":<{stroke_count_width}}')
        for d, num_strokes in date_to_stroke_stats_list:
            lines.append(f'{d:<13} {num_strokes:<{stroke_count_width}}')
        return lines


##############################
# States

class State:
    def __init__(self, engine, listener_manager):
        self.engine = engine
        self.listener_manager = listener_manager

    def load(self):
        raise NotImplementedError


class WelcomeToNinjaTraining(State):
    def load(self):
        phrase = ['i', 'am', 'ready', 'to', 'practice', '', '', '']
        callback = GreetingCallback(self.engine, mini_lesson=True)
        wait_listener = WaitForPhrase(phrase, callback)

        self.listener_manager.add_listener('welcome_state.interactive_session_requested',
                                           wait_listener)

        phrase = ['i', 'am', 'ready', 'to', 'practice', 'with', 'data', '', '', '']
        callback = GreetingCallback(self.engine, mini_lesson=False)
        wait_listener = WaitForPhrase(phrase, callback)

        self.listener_manager.add_listener('welcome_state.interactive_session_requested_with_details',
                                           wait_listener)

        phrase = ['show', 'stats', '', '', '']
        callback = StatsCallback(self.engine, days=10)
        wait_listener = WaitForPhrase(phrase, callback)

        self.listener_manager.add_listener('welcome_state.show_stats_for_ten_days',
                                           wait_listener)

        phrase = ['show', 'week', '', '', '']
        callback = StatsCallback(self.engine, days=7)
        wait_listener = WaitForPhrase(phrase, callback)

        self.listener_manager.add_listener('welcome_state.show_stats_for_week',
                                           wait_listener)

        phrase = ['show', 'month', '', '', '']
        callback = StatsCallback(self.engine, days=30)
        wait_listener = WaitForPhrase(phrase, callback)

        self.listener_manager.add_listener('welcome_state.show_stats_for_month',
                                           wait_listener)



#!/usr/bin/env python3

from pathlib import Path
from datetime import date

from plover_ninja.storage import get_slowest_stroked_words, TEMP_DIR


class NeedForSpeed:
    def __init__(self):
        pass

    # TODO: Make this a class method?
    def make_lesson(self, mini_lesson=False):
        # Returns tuples containing:
        # - word_id
        # - word
        # - practice weight
        # - average duration
        # - weighted duration
        average_stroke_duration_tuples = get_slowest_stroked_words()

        if len(average_stroke_duration_tuples) == 0:
            return 'Cannot recommend any words for speedstudy; \n' + \
                   'no words have been written yet!'

        # TODO: make this a util function
        longest_word_len = 0
        for _, word, _, _, _ in average_stroke_duration_tuples:
            word_len = len(word)
            if word_len > longest_word_len:
                longest_word_len = word_len
        word_padding = longest_word_len + 5

        # TODO: make this a util function
        longest_word_id_len = 0
        for word_id, _, _, _, _ in average_stroke_duration_tuples:
            word_id_len = len(str(word_id))
            if word_id_len > longest_word_id_len:
                longest_word_id_len = word_id_len
        word_id_padding = longest_word_id_len + 5

        word_list = []
        detailed_word_list = []
        for word_id, word, practice_weight, average_duration, weighted_duration in average_stroke_duration_tuples:
            text = f'{word:{word_padding}} word #{word_id:<{word_id_padding}} avg stroke duration: {average_duration:.3f}  weighted duration: {weighted_duration:.3f}'
            detailed_word_list.append(text)
            word_list.append(word)
        detailed_word_list = '\n'.join(detailed_word_list)
        word_list = '\n'.join(word_list)

        today = date.today()
        file_name_today = today.strftime('%Y%m%d')
        report_text_today = today.strftime('%m/%d/%Y')

        with open(f'{Path(TEMP_DIR) / file_name_today}_need_for_speed.txt', 'w', encoding='utf-8') as f:
            text = f"""Ninja Lesson
{report_text_today}

Take these words out for some speed practice in order to boost
your overall efficiency. Here are some tips for speed practice:

- Consider using a metronome (digital, physical, whatever is handy)
- Find a starting tempo where you feel _absolutely comfortable_
  stenoing the words. This may feel incredibly slow to you. That's okay!
- Very slowly, bump up the speed as you start to feel more comfortable
  with the words. Pushing a bit beyond your edge is good. Try to
  avoid straining too much though. Bring the tempo back down as
  needed.
- While you practice, look for 'speed bumps'. A speed bump may be:
  - a mental pause to remember a translation
  - a mental pause while you try to recall a handshape
  - a physical delay while you try to make the shape of a chord
  - a physical delay while you transition between chords
- Slowly work with each speed bump you notice, giving it the
  attention it needs until you can feel the mental or physical
  hesitation start to relax, and you can feel that you can stroke
  the word with greater ease.
- Be patient with yourself! Smoothing out speed bumps is about slow,
  careful work. Let speed be a byproduct of this careful work.

You can do it!!!

{detailed_word_list}

🐦🥋
"""
            f.write(text)

        if mini_lesson:
            text = f"""Need For Speed Words!
---------------------
{word_list}
"""
            return text
        return text



from plover_ninja.plugins.ninja_test import NinjaTest
from plover_ninja.plugins.ninja_repl import NinjaRepl
from plover_ninja.plugins.stroke_efficiency_log import StrokeEfficiencyLog

import logging
logger = logging.getLogger(__name__)

NINJA_PLUGINS = [NinjaRepl, StrokeEfficiencyLog]


class Main:
    def __init__(self, engine):
        self.engine = engine
        self.ninja_plugins = [cls(engine) for cls in NINJA_PLUGINS]


    def start(self):
        self.engine.hook_connect('stroked', self._on_stroked)
        self.engine.hook_connect('translated', self._on_translated)

    def stop(self):
        self.engine.hook_disconnect('stroked', self._on_stroked)
        self.engine.hook_disconnect('translated', self._on_translated)

    def _on_translated(self, old, new):
        for plugin in self.ninja_plugins:
            plugin.on_translated(old, new)

    def _on_stroked(self, stroke):
        for plugin in self.ninja_plugins:
            plugin.on_stroked(stroke)
